"""Sparse DIAgonal format"""

__docformat__ = "restructuredtext en"

__all__ = ['dia_array', 'dia_matrix', 'isspmatrix_dia']

import numpy as np

from .._lib._util import _prune_array, copy_if_needed
from ._matrix import spmatrix
from ._base import issparse, _formats, _spbase, sparray
from ._data import _data_matrix
from ._sputils import (
    isdense, isscalarlike, isshape, upcast_char, getdtype, get_sum_dtype,
    validateaxis, check_shape
)
from ._sparsetools import dia_matmat, dia_matvec, dia_matvecs, dia_tocsr


class _dia_base(_data_matrix):
    _format = 'dia'

    def __init__(self, arg1, shape=None, dtype=None, copy=False, *, maxprint=None):
        _data_matrix.__init__(self, arg1, maxprint=maxprint)

        if issparse(arg1):
            if arg1.format == "dia":
                if copy:
                    arg1 = arg1.copy()
                self.data = arg1.data
                self.offsets = arg1.offsets
                self._shape = check_shape(arg1.shape)
            else:
                if arg1.format == self.format and copy:
                    A = arg1.copy()
                else:
                    A = arg1.todia()
                self.data = A.data
                self.offsets = A.offsets
                self._shape = check_shape(A.shape)
        elif isinstance(arg1, tuple):
            if isshape(arg1):
                # It's a tuple of matrix dimensions (M, N)
                # create empty matrix
                self._shape = check_shape(arg1)
                self.data = np.zeros((0,0), getdtype(dtype, default=float))
                idx_dtype = self._get_index_dtype(maxval=max(self.shape))
                self.offsets = np.zeros((0), dtype=idx_dtype)
            else:
                try:
                    # Try interpreting it as (data, offsets)
                    data, offsets = arg1
                except Exception as e:
                    message = 'unrecognized form for dia_array constructor'
                    raise ValueError(message) from e
                else:
                    if shape is None:
                        raise ValueError('expected a shape argument')
                    if not copy:
                        copy = copy_if_needed
                    self.data = np.atleast_2d(np.array(arg1[0], dtype=dtype, copy=copy))
                    offsets = np.array(arg1[1],
                                       dtype=self._get_index_dtype(maxval=max(shape)),
                                       copy=copy)
                    self.offsets = np.atleast_1d(offsets)
                    self._shape = check_shape(shape)
        else:
            # must be dense, convert to COO first, then to DIA
            try:
                arg1 = np.asarray(arg1)
            except Exception as e:
                raise ValueError("unrecognized form for "
                                 f"{self.format}_matrix constructor") from e
            if isinstance(self, sparray) and arg1.ndim != 2:
                raise ValueError(f"DIA arrays don't support {arg1.ndim}D input. Use 2D")
            A = self._coo_container(arg1, dtype=dtype, shape=shape).todia()
            self.data = A.data
            self.offsets = A.offsets
            self._shape = check_shape(A.shape)

        if dtype is not None:
            newdtype = getdtype(dtype)
            self.data = self.data.astype(newdtype, copy=False)

        # check format
        if self.offsets.ndim != 1:
            raise ValueError('offsets array must have rank 1')

        if self.data.ndim != 2:
            raise ValueError('data array must have rank 2')

        if self.data.shape[0] != len(self.offsets):
            raise ValueError(
                f'number of diagonals ({self.data.shape[0]}) does not match the number '
                f'of offsets ({len(self.offsets)})'
            )
        if len(np.unique(self.offsets)) != len(self.offsets):
            raise ValueError('offset array contains duplicate values')

    def __repr__(self):
        _, fmt = _formats[self.format]
        sparse_cls = 'array' if isinstance(self, sparray) else 'matrix'
        d = self.data.shape[0]
        return (
            f"<{fmt} sparse {sparse_cls} of dtype '{self.dtype}'\n"
            f"\twith {self.nnz} stored elements ({d} diagonals) and shape {self.shape}>"
        )

    def _data_mask(self):
        """Returns a mask of the same shape as self.data, where
        mask[i,j] is True when data[i,j] corresponds to a stored element."""
        num_rows, num_cols = self.shape
        offset_inds = np.arange(self.data.shape[1])
        row = offset_inds - self.offsets[:,None]
        mask = (row >= 0)
        mask &= (row < num_rows)
        mask &= (offset_inds < num_cols)
        return mask

    def count_nonzero(self, axis=None):
        if axis is not None:
            raise NotImplementedError(
                "count_nonzero over an axis is not implemented for DIA format"
            )
        mask = self._data_mask()
        return np.count_nonzero(self.data[mask])

    count_nonzero.__doc__ = _spbase.count_nonzero.__doc__

    def _getnnz(self, axis=None):
        if axis is not None:
            raise NotImplementedError("_getnnz over an axis is not implemented "
                                      "for DIA format")
        M, N = self.shape
        L = min(self.data.shape[1], N)
        return int(np.maximum(np.minimum(M + self.offsets, L) -
                              np.maximum(self.offsets, 0),
                              0).sum())

    _getnnz.__doc__ = _spbase._getnnz.__doc__

    def sum(self, axis=None, dtype=None, out=None):
        axis = validateaxis(axis)

        res_dtype = dtype if dtype is not None else get_sum_dtype(self.dtype)
        num_rows, num_cols = self.shape
        ret = None

        if axis == (0,):
            mask = self._data_mask()
            x = (self.data * mask).sum(dtype=res_dtype, axis=0)
            if x.shape[0] == num_cols:
                res = x
            else:
                res = np.zeros(num_cols, dtype=x.dtype)
                res[:x.shape[0]] = x
            ret = self._ascontainer(res, dtype=res_dtype)

        else:  # axis is None or (1,)
            row_sums = np.zeros((num_rows, 1), dtype=res_dtype)
            one = np.ones(num_cols, dtype=res_dtype)
            dia_matvec(num_rows, num_cols, len(self.offsets),
                       self.data.shape[1], self.offsets, 
                       self.data.astype(res_dtype), one, row_sums)

            row_sums = self._ascontainer(row_sums)

            if axis is None:
                return row_sums.sum(dtype=dtype, out=out)

            ret = self._ascontainer(row_sums.sum(axis=axis))

        return ret.sum(axis=(), dtype=dtype, out=out)

    sum.__doc__ = _spbase.sum.__doc__

    def _add_sparse(self, other, sub=False):
        # If other is not DIA format, let them handle us instead.
        if not isinstance(other, _dia_base):
            return other._add_sparse(self)

        # Fast path for exact equality of the sparsity structure.
        if np.array_equal(self.offsets, other.offsets):
            return self._with_data(self.data - other.data if sub else
                                   self.data + other.data)

        # Find the union of the offsets (which will be sorted and unique).
        new_offsets = np.union1d(self.offsets, other.offsets)
        self_idx = np.searchsorted(new_offsets, self.offsets)
        other_idx = np.searchsorted(new_offsets, other.offsets)

        self_d = self.data.shape[1]
        other_d = other.data.shape[1]
        # Fast path for a sparsity structure where the final offsets are a
        # permutation of the existing offsets and the diagonal lengths match.
        if self_d == other_d and len(new_offsets) == len(self.offsets):
            new_data = self.data[_invert_index(self_idx)]
            if sub:
                new_data[other_idx, :] -= other.data
            else:
                new_data[other_idx, :] += other.data
        elif self_d == other_d and len(new_offsets) == len(other.offsets):
            if sub:
                new_data = -other.data[_invert_index(other_idx)]
            else:
                new_data = other.data[_invert_index(other_idx)]
            new_data[self_idx, :] += self.data
        else:
            # Maximum diagonal length of the result.
            d = min(self.shape[0] + new_offsets[-1], self.shape[1])

            # Add all diagonals to a freshly allocated data array.
            new_data = np.zeros(
                (len(new_offsets), d),
                dtype=np.result_type(self.data, other.data),
            )
            new_data[self_idx, :self_d] += self.data[:, :d]
            if sub:
                new_data[other_idx, :other_d] -= other.data[:, :d]
            else:
                new_data[other_idx, :other_d] += other.data[:, :d]
        return self._dia_container((new_data, new_offsets), shape=self.shape)

    def _sub_sparse(self, other):
        # If other is not DIA format, use default handler.
        if not isinstance(other, _dia_base):
            return super()._sub_sparse(other)

        return self._add_sparse(other, sub=True)

    def _mul_scalar(self, other):
        return self._with_data(self.data * other)

    def multiply(self, other):
        if isscalarlike(other):
            return self._mul_scalar(other)

        if isdense(other):
            if other.ndim > 2:
                return self.toarray() * other

            # Use default handler for pathological cases.
            if 0 in self.shape or 1 in self.shape or 0 in other.shape:
                return super().multiply(other)

            other = np.atleast_2d(other)
            other_rows, other_cols = other.shape
            rows, cols = self.shape
            L = min(self.data.shape[1], cols)
            data = self.data[:, :L].astype(np.result_type(self.data, other))  # copy
            if other_rows == 1:
                data *= other[0, :L]
            elif other_rows != rows:
                raise ValueError('inconsistent shapes')
            else:
                j = np.arange(L)
                if L > rows:
                    i = (j - self.offsets[:, None]) % rows
                else:  # can use faster method
                    i = j - self.offsets[:, None] % rows
                if other_cols == 1:
                    j = 0
                elif other_cols != cols:
                    raise ValueError('inconsistent shapes')
                data *= other[i, j]
            return self._with_data(data)

        # If other is not DIA format or needs broadcasting (unreasonable
        # use case for DIA anyway), use default handler.
        if not isinstance(other, _dia_base) or other.shape != self.shape:
            return super().multiply(other)

        # Find common offsets (unique diagonals don't contribute)
        # and indices corresponding to them in multiplicand and multiplier.
        offsets, self_idx, other_idx = \
            np.intersect1d(self.offsets, other.offsets,
                           assume_unique=True, return_indices=True)
        # Only overlapping length of diagonals can have non-zero products.
        L = min(self.data.shape[1], other.data.shape[1])
        data = self.data[self_idx, :L] * other.data[other_idx, :L]
        return self._dia_container((data, offsets), shape=self.shape)

    multiply.__doc__ = _data_matrix.multiply.__doc__

    def _matmul_vector(self, other):
        x = other

        y = np.zeros(self.shape[0], dtype=upcast_char(self.dtype.char,
                                                       x.dtype.char))

        L = self.data.shape[1]

        M,N = self.shape

        dia_matvec(M,N, len(self.offsets), L, self.offsets, self.data,
                   x.ravel(), y.ravel())

        return y

    def _matmul_multivector(self, other):
        res = np.zeros((self.shape[0], other.shape[1]),
                       dtype=np.result_type(self.data, other))
        dia_matvecs(*self.shape, *self.data.shape, self.offsets, self.data,
                    other.shape[1], other, res)
        return res

    def _matmul_sparse(self, other):
        # If other is not DIA format, use default handler.
        if not isinstance(other, _dia_base):
            return super()._matmul_sparse(other)

        # If any dimension is zero, return empty array immediately.
        if 0 in self.shape or 0 in other.shape:
            return self._dia_container((self.shape[0], other.shape[1]))

        offsets, data = dia_matmat(*self.shape, *self.data.shape,
                                   self.offsets, self.data,
                                   other.shape[1], *other.data.shape,
                                   other.offsets, other.data)
        return self._dia_container((data.reshape(len(offsets), -1), offsets),
                                   (self.shape[0], other.shape[1]))

    def _setdiag(self, values, k=0):
        M, N = self.shape

        if values.ndim == 0:
            # broadcast
            values_n = np.inf
        else:
            values_n = len(values)

        if k < 0:
            n = min(M + k, N, values_n)
            min_index = 0
            max_index = n
        else:
            n = min(M, N - k, values_n)
            min_index = k
            max_index = k + n

        if values.ndim != 0:
            # allow also longer sequences
            values = values[:n]

        data_rows, data_cols = self.data.shape
        if k in self.offsets:
            if max_index > data_cols:
                data = np.zeros((data_rows, max_index), dtype=self.data.dtype)
                data[:, :data_cols] = self.data
                self.data = data
            self.data[self.offsets == k, min_index:max_index] = values
        else:
            self.offsets = np.append(self.offsets, self.offsets.dtype.type(k))
            m = max(max_index, data_cols)
            data = np.zeros((data_rows + 1, m), dtype=self.data.dtype)
            data[:-1, :data_cols] = self.data
            data[-1, min_index:max_index] = values
            self.data = data

    def todia(self, copy=False):
        if copy:
            return self.copy()
        else:
            return self

    todia.__doc__ = _spbase.todia.__doc__

    def transpose(self, axes=None, copy=False):
        if axes is not None and axes != (1, 0):
            raise ValueError("Sparse arrays/matrices do not support "
                              "an 'axes' parameter because swapping "
                              "dimensions is the only logical permutation.")

        num_rows, num_cols = self.shape
        max_dim = max(self.shape)

        # flip diagonal offsets
        offsets = -self.offsets

        # re-align the data matrix
        r = np.arange(len(offsets), dtype=np.intc)[:, None]
        c = np.arange(num_rows, dtype=np.intc) - (offsets % max_dim)[:, None]
        pad_amount = max(0, max_dim-self.data.shape[1])
        data = np.hstack((self.data, np.zeros((self.data.shape[0], pad_amount),
                                              dtype=self.data.dtype)))
        data = data[r, c]
        return self._dia_container((data, offsets), shape=(
            num_cols, num_rows), copy=copy)

    transpose.__doc__ = _spbase.transpose.__doc__

    def diagonal(self, k=0):
        rows, cols = self.shape
        if k <= -rows or k >= cols:
            return np.empty(0, dtype=self.data.dtype)
        idx, = np.nonzero(self.offsets == k)
        first_col = max(0, k)
        last_col = min(rows + k, cols)
        result_size = last_col - first_col
        if idx.size == 0:
            return np.zeros(result_size, dtype=self.data.dtype)
        result = self.data[idx[0], first_col:last_col]
        padding = result_size - len(result)
        if padding > 0:
            result = np.pad(result, (0, padding), mode='constant')
        return result

    diagonal.__doc__ = _spbase.diagonal.__doc__

    def tocsr(self, copy=False):
        if 0 in self.shape or len(self.offsets) == 0:
            return self._csr_container(self.shape, dtype=self.dtype)

        n_rows, n_cols = self.shape
        max_nnz = self.nnz
        # np.argsort always returns dtype=int, which can cause automatic dtype
        # expansion for everything else even if not needed (see gh19245), but
        # CSR wants common dtype for indices, indptr and shape, so care should
        # be taken to use appropriate indexing dtype throughout.
        idx_dtype = self._get_index_dtype(maxval=max(max_nnz, n_rows, n_cols))
        order = np.argsort(self.offsets).astype(idx_dtype, copy=False)
        csr_data = np.empty(max_nnz, dtype=self.dtype)
        indices = np.empty(max_nnz, dtype=idx_dtype)
        indptr = np.empty(1 + n_rows, dtype=idx_dtype)
        # Conversion eliminates explicit zeros and returns actual nnz.
        nnz = dia_tocsr(n_rows, n_cols, *self.data.shape,
                        self.offsets.astype(idx_dtype, copy=False), self.data,
                        order, csr_data, indices, indptr)
        # Shrink indexing dtype, if needed, and prune arrays.
        idx_dtype = self._get_index_dtype(maxval=max(nnz, n_rows, n_cols))
        csr_data = _prune_array(csr_data[:nnz])
        indices = _prune_array(indices[:nnz].astype(idx_dtype, copy=False))
        indptr = indptr.astype(idx_dtype, copy=False)
        out = self._csr_container((csr_data, indices, indptr),
                                  shape=self.shape, dtype=self.dtype)
        out.has_canonical_format = True
        return out

    tocsr.__doc__ = _spbase.tocsr.__doc__

    # needed by _data_matrix
    def _with_data(self, data, copy=True):
        """Returns a matrix with the same sparsity structure as self,
        but with different data.  By default the structure arrays are copied.
        """
        if copy:
            return self._dia_container(
                (data, self.offsets.copy()), shape=self.shape
            )
        else:
            return self._dia_container(
                (data, self.offsets), shape=self.shape
            )

    def resize(self, *shape):
        shape = check_shape(shape)
        M, N = shape
        # we do not need to handle the case of expanding N
        self.data = self.data[:, :N]

        if (M > self.shape[0] and
                np.any(self.offsets + self.shape[0] < self.data.shape[1])):
            # explicitly clear values that were previously hidden
            mask = (self.offsets[:, None] + self.shape[0] <=
                    np.arange(self.data.shape[1]))
            self.data[mask] = 0

        self._shape = shape

    resize.__doc__ = _spbase.resize.__doc__


def _invert_index(idx):
    """Helper function to invert an index array."""
    inv = np.zeros_like(idx)
    inv[idx] = np.arange(len(idx))
    return inv


def isspmatrix_dia(x):
    """Is `x` of dia_matrix type?

    .. warning::

       SciPy sparse is shifting from a sparse matrix interface to a sparse
       array interface. In the next few releases we expect to deprecate the
       sparse matrix interface. For documentation of the matrix
       interface, see the :ref:`spmatrix interface docs <spmatrix_api>`.
       For guidance on converting existing code to sparse arrays, see
       :ref:`Migration from spmatrix to sparray <migration_to_sparray>`.

    Parameters
    ----------
    x
        object to check for being a dia matrix

    Returns
    -------
    bool
        True if `x` is a dia matrix, False otherwise

    Examples
    --------
    >>> from scipy.sparse import dia_array, dia_matrix, coo_matrix, isspmatrix_dia
    >>> isspmatrix_dia(dia_matrix([[5]]))
    True
    >>> isspmatrix_dia(dia_array([[5]]))
    False
    >>> isspmatrix_dia(coo_matrix([[5]]))
    False
    """
    return isinstance(x, dia_matrix)


# This namespace class separates array from matrix with isinstance
class dia_array(_dia_base, sparray):
    """
    Sparse array with DIAgonal storage.

    This can be instantiated in several ways:
        dia_array(D)
            where D is a 2-D ndarray

        dia_array(S)
            with another sparse array or matrix S (equivalent to S.todia())

        dia_array((M, N), [dtype])
            to construct an empty array with shape (M, N),
            dtype is optional, defaulting to dtype='d'.

        dia_array((data, offsets), shape=(M, N))
            where the ``data[k,:]`` stores the diagonal entries for
            diagonal ``offsets[k]`` (See example below)

    Attributes
    ----------
    data
        DIA format data array of the array
    offsets
        DIA format offset array of the array
    dtype : dtype
        Data type of the array
    shape : 2-tuple
        Shape of the array
    ndim : int
        Number of dimensions (this is always 2)
    format : str
        Three letter code for the format of the array storage, e.g. 'dia'
    nnz : int
        Number of values stored in the array
    size : int
        Number of values stored in the array
    T : dia_array
        The transpose of the array
    mT : dia_array
        The matrix transpose of the array

    Notes
    -----

    Sparse arrays can be used in arithmetic operations: they support
    addition, subtraction, multiplication, division, and matrix power.
    Sparse arrays with DIAgonal storage do not support slicing.

    **Format details**

    The ``data`` array stores the diagonal elements. The alignment of these
    elements within the rows of ``data`` depends on their position relative
    to the main diagonal:

    * **Main diagonal** (``offsets[i] == 0``): Elements start at column 0.
    * **Super-diagonals** (``offsets[i] > 0``): Elements are right-aligned
      (padded with zeros on the left).
    * **Sub-diagonals** (``offsets[i] < 0``): Elements are left-aligned
      (padded with zeros on the right).

    Each column of ``data`` corresponds to a diagonal in the resulting matrix.

    Mathematically, the element at row ``r`` and column ``c`` of the matrix is
    stored in the ``data`` array at row ``i`` and column
    ``c - max(0, -offsets[i])``, where ``i`` is the index of the diagonal
    in ``offsets``.

    Note that if ``offsets`` is provided in decreasing order, this format
    matches the BLAS/LAPACK general band format (e.g., as used in ``dgbmv``).

    Examples
    --------

    >>> import numpy as np
    >>> from scipy.sparse import dia_array
    >>> dia_array((3, 4), dtype=np.int8).toarray()
    array([[0, 0, 0, 0],
           [0, 0, 0, 0],
           [0, 0, 0, 0]], dtype=int8)

    >>> data = np.array([[1, 2, 3, 4]]).repeat(3, axis=0)
    >>> offsets = np.array([0, -1, 2])
    >>> dia_array((data, offsets), shape=(4, 4)).toarray()
    array([[1, 0, 3, 0],
           [1, 2, 0, 4],
           [0, 2, 3, 0],
           [0, 0, 3, 4]])

    >>> from scipy.sparse import dia_array
    >>> n = 10
    >>> ex = np.ones(n)
    >>> data = np.array([ex, 2 * ex, ex])
    >>> offsets = np.array([-1, 0, 1])
    >>> dia_array((data, offsets), shape=(n, n)).toarray()
    array([[2., 1., 0., ..., 0., 0., 0.],
           [1., 2., 1., ..., 0., 0., 0.],
           [0., 1., 2., ..., 0., 0., 0.],
           ...,
           [0., 0., 0., ..., 2., 1., 0.],
           [0., 0., 0., ..., 1., 2., 1.],
           [0., 0., 0., ..., 0., 1., 2.]])
    """  # numpydoc ignore=PR01


class dia_matrix(spmatrix, _dia_base):
    """
    Sparse matrix with DIAgonal storage.

    .. warning::

       SciPy sparse is shifting from a sparse matrix interface to a sparse
       array interface. In the next few releases we expect to deprecate the
       sparse matrix interface. For documentation of the matrix
       interface, see the :ref:`spmatrix interface docs <spmatrix_api>`.
       For guidance on converting existing code to sparse arrays, see
       :ref:`Migration from spmatrix to sparray <migration_to_sparray>`.

    This can be instantiated in several ways:
        dia_matrix(D)
            where D is a 2-D ndarray

        dia_matrix(S)
            with another sparse array or matrix S (equivalent to S.todia())

        dia_matrix((M, N), [dtype])
            to construct an empty matrix with shape (M, N),
            dtype is optional, defaulting to dtype='d'.

        dia_matrix((data, offsets), shape=(M, N))
            where the ``data[k,:]`` stores the diagonal entries for
            diagonal ``offsets[k]`` (See example below)

    Attributes
    ----------
    data
        DIA format data array of the matrix
    offsets
        DIA format offset array of the matrix
    dtype : dtype
        Data type of the matrix
    shape : 2-tuple
        Shape of the matrix
    ndim : int
        Number of dimensions (this is always 2)
    format : str
        Three letter code for the format of the matrix storage, e.g. 'dia'
    nnz : int
        Number of values stored in the matrix
    size : int
        Number of values stored in the matrix
    T : dia_matrix
        The transpose of the matrix
    mT : dia_matrix
        The matrix transpose

    Notes
    -----
    Sparse matrices can be used in arithmetic operations: they support
    addition, subtraction, multiplication, division, and matrix power.
    Sparse arrays with DIAgonal storage do not support slicing.

    **Format details**

    The ``data`` array stores the diagonal elements. The alignment of these
    elements within the rows of ``data`` depends on their position relative
    to the main diagonal:

    * **Main diagonal** (``offsets[i] == 0``): Elements start at column 0.
    * **Super-diagonals** (``offsets[i] > 0``): Elements are right-aligned
      (padded with zeros on the left).
    * **Sub-diagonals** (``offsets[i] < 0``): Elements are left-aligned
      (padded with zeros on the right).

    Each column of ``data`` corresponds to a diagonal in the resulting matrix.

    Mathematically, the element at row ``r`` and column ``c`` of the matrix is
    stored in the ``data`` array at row ``i`` and column
    ``c - max(0, -offsets[i])``, where ``i`` is the index of the diagonal
    in ``offsets``.

    Note that if ``offsets`` is provided in decreasing order, this format
    matches the BLAS/LAPACK general band format (e.g., as used in ``dgbmv``).

    Examples
    --------

    >>> import numpy as np
    >>> from scipy.sparse import dia_matrix
    >>> dia_matrix((3, 4), dtype=np.int8).toarray()
    array([[0, 0, 0, 0],
           [0, 0, 0, 0],
           [0, 0, 0, 0]], dtype=int8)

    >>> data = np.array([[1, 2, 3, 4]]).repeat(3, axis=0)
    >>> offsets = np.array([0, -1, 2])
    >>> dia_matrix((data, offsets), shape=(4, 4)).toarray()
    array([[1, 0, 3, 0],
           [1, 2, 0, 4],
           [0, 2, 3, 0],
           [0, 0, 3, 4]])

    >>> from scipy.sparse import dia_matrix
    >>> n = 10
    >>> ex = np.ones(n)
    >>> data = np.array([ex, 2 * ex, ex])
    >>> offsets = np.array([-1, 0, 1])
    >>> dia_matrix((data, offsets), shape=(n, n)).toarray()
    array([[2., 1., 0., ..., 0., 0., 0.],
           [1., 2., 1., ..., 0., 0., 0.],
           [0., 1., 2., ..., 0., 0., 0.],
           ...,
           [0., 0., 0., ..., 2., 1., 0.],
           [0., 0., 0., ..., 1., 2., 1.],
           [0., 0., 0., ..., 0., 1., 2.]])
    """



"""Static type stubs for `_compat.py`."""

# https://github.com/scikit-learn/scikit-learn/pull/27910#issuecomment-2568023972
from __future__ import annotations

from types import ModuleType
from typing import Any, TypeGuard

# TODO import from typing (requires Python >=3.13)
from typing_extensions import TypeIs

from ._typing import Array, Device

# pylint: disable=missing-class-docstring,unused-argument

def array_namespace(
    *xs: Array | complex | None,
    api_version: str | None = None,
    use_compat: bool | None = None,
) -> ModuleType: ...
def device(x: Array, /) -> Device: ...
def is_array_api_obj(x: object, /) -> TypeIs[Array]: ...
def is_array_api_strict_namespace(xp: ModuleType, /) -> bool: ...
def is_cupy_namespace(xp: ModuleType, /) -> bool: ...
def is_dask_namespace(xp: ModuleType, /) -> bool: ...
def is_jax_namespace(xp: ModuleType, /) -> bool: ...
def is_numpy_namespace(xp: ModuleType, /) -> bool: ...
def is_pydata_sparse_namespace(xp: ModuleType, /) -> bool: ...
def is_torch_namespace(xp: ModuleType, /) -> bool: ...
def is_cupy_array(x: object, /) -> TypeGuard[Array]: ...
def is_dask_array(x: object, /) -> TypeGuard[Array]: ...
def is_jax_array(x: object, /) -> TypeGuard[Array]: ...
def is_numpy_array(x: object, /) -> TypeGuard[Array]: ...
def is_pydata_sparse_array(x: object, /) -> TypeGuard[Array]: ...
def is_torch_array(x: object, /) -> TypeGuard[Array]: ...
def is_lazy_array(x: object, /) -> TypeGuard[Array]: ...
def is_writeable_array(x: object, /) -> TypeGuard[Array]: ...
def size(x: Array, /) -> int | None: ...
def to_device(
    x: Array,
    device: Device,  # pylint: disable=redefined-outer-name
    /,
    *,
    stream: int | Any | None = None,
) -> Array: ...



"""Test of csgraph public API with int64 index arrays in csr format.

See gh-24629
"""
import numpy as np
import pytest

from scipy.sparse import csr_array
import scipy.sparse.csgraph as csgraph_subpackage
from scipy.sparse.csgraph import (
    connected_components,
    laplacian,
    shortest_path,
    floyd_warshall,
    dijkstra,
    bellman_ford,
    johnson,
    yen,
    breadth_first_order,
    depth_first_order,
    breadth_first_tree,
    depth_first_tree,
    minimum_spanning_tree,
    reverse_cuthill_mckee,
    maximum_flow,
    maximum_bipartite_matching,
    min_weight_full_bipartite_matching,
    structural_rank,
    construct_dist_matrix,
    reconstruct_path,
    csgraph_to_dense,
    csgraph_to_masked,
)


@pytest.fixture()
def A():
    A = csr_array([[0, 1, 2], [2, 1, 0], [0, 1, 0]])
    A.indptr = A.indptr.astype(np.int64)
    A.indices = A.indices.astype(np.int64)
    return A


check_these_functions = [
    # (csgraph_function, args),
    (connected_components, ()),
    (laplacian, ()),
    (floyd_warshall, ()),
    (dijkstra, ()),
    (bellman_ford, ()),
    (johnson, ()),
    (yen, (0, 1, 1)),
    (breadth_first_order, (0,)),
    (depth_first_order, (0,)),
    (breadth_first_tree, (0,)),
    (depth_first_tree, (0,)),
    (minimum_spanning_tree, ()),
    (reverse_cuthill_mckee, ()),
    (maximum_flow, (0, 1)),
    (maximum_bipartite_matching, ()),
    (min_weight_full_bipartite_matching, ()),
    (structural_rank, ()),
    (csgraph_to_dense, ()),
    (csgraph_to_masked, ()),
]


def test_new_functions_untested_here():
    #This tests whether any new code has arrived that isn't in the list above.

    # funcs tested individually outside of check_these_functions
    handled = ["shortest_path", "construct_dist_matrix", "reconstruct_path",
               "csgraph_from_dense", "csgraph_from_masked", "csgraph_masked_from_dense"]

    # look for functions that should be tested
    functions = [fname for fname in dir(csgraph_subpackage)
                 if fname[0] != "_"
                 if not fname.endswith("Error")
                 if not fname.startswith("test")
                 if fname not in handled]

    checked_funcs = [x[0].__name__ for x in check_these_functions]
    for func in functions:
        if func not in checked_funcs:
            msg = (f'function "{func}" in csgraph is not tested. '
                   'It should be added to the `check_these_functions` list above or '
                   'handled separately if its signature has many required parameters')
            raise ValueError(msg)


@pytest.mark.parametrize("csgraph_function, args", check_these_functions)
def test_smoke_from_int64_index_arrays(csgraph_function, args, A):
    csgraph_function(A, *args)


def test_smoke_from_int64_construct_dist_matrix(A):
    _, preds = shortest_path(A, return_predecessors=True)
    construct_dist_matrix(A, preds)


def test_smoke_from_int64_shortest_path(A):
    shortest_path(A, method='auto')


def test_smoke_from_int64_reconstruct_path(A):
    pred = np.array([-9999, 0, 1])
    reconstruct_path(A, pred)




"""
Module to expose more detailed version info for the installed `scipy`
"""
version = "1.18.0"
full_version = version
short_version = version.split('.dev')[0]
git_revision = "54ef5423f2e4376230ec3bfda6912a07a50958e3"
release = 'dev' not in version and '+' not in version

if not release:
    version = full_version



from typing import Any, Literal, assert_type

import numpy as np
import numpy.typing as npt
from numpy.linalg._linalg import (
    EighResult,
    EigResult,
    QRResult,
    SlogdetResult,
    SVDResult,
)

type _Array1D[ScalarT: np.generic] = np.ndarray[tuple[int], np.dtype[ScalarT]]
type _Array2D[ScalarT: np.generic] = np.ndarray[tuple[int, int], np.dtype[ScalarT]]
type _Array3D[ScalarT: np.generic] = np.ndarray[tuple[int, int, int], np.dtype[ScalarT]]

bool_list_1d: list[bool]
bool_list_2d: list[list[bool]]
int_list_1d: list[int]
int_list_2d: list[list[int]]
float_list_1d: list[float]
float_list_2d: list[list[float]]
float_list_3d: list[list[list[float]]]
float_list_4d: list[list[list[list[float]]]]
complex_list_1d: list[complex]
complex_list_2d: list[list[complex]]
complex_list_3d: list[list[list[complex]]]
bytes_list_2d: list[list[bytes]]
str_list_2d: list[list[str]]

AR_any: np.ndarray
AR_f_: npt.NDArray[np.floating]
AR_c_: npt.NDArray[np.complexfloating]
AR_i8: npt.NDArray[np.int64]
AR_f2: npt.NDArray[np.float16]
AR_f4: npt.NDArray[np.float32]
AR_f8: npt.NDArray[np.float64]
AR_f10: npt.NDArray[np.longdouble]
AR_c8: npt.NDArray[np.complex64]
AR_c16: npt.NDArray[np.complex128]
AR_c20: npt.NDArray[np.clongdouble]
AR_O: npt.NDArray[np.object_]
AR_M: npt.NDArray[np.datetime64]
AR_m: npt.NDArray[np.timedelta64]
AR_S: npt.NDArray[np.bytes_]
AR_U: npt.NDArray[np.str_]
AR_b: npt.NDArray[np.bool]

AR_b_1d: np.ndarray[tuple[int], np.dtype[np.bool]]
AR_b_2d: np.ndarray[tuple[int, int], np.dtype[np.bool]]

AR_i8_1d: np.ndarray[tuple[int], np.dtype[np.int64]]
AR_i8_2d: np.ndarray[tuple[int, int], np.dtype[np.int64]]

SC_f8: np.float64
AR_f8_0d: np.ndarray[tuple[()], np.dtype[np.float64]]
AR_f8_1d: _Array1D[np.float64]
AR_f8_2d: _Array2D[np.float64]
AR_f8_3d: _Array3D[np.float64]
AR_f8_4d: np.ndarray[tuple[int, int, int, int], np.dtype[np.float64]]

AR_f2_2d: _Array2D[np.float16]
AR_f4_1d: _Array1D[np.float32]
AR_f4_2d: _Array2D[np.float32]
AR_f4_3d: _Array3D[np.float32]
AR_f10_2d: _Array2D[np.longdouble]
AR_f10_3d: _Array3D[np.longdouble]

AR_c16_1d: np.ndarray[tuple[int], np.dtype[np.complex128]]
AR_c16_2d: np.ndarray[tuple[int, int], np.dtype[np.complex128]]

###

assert_type(np.linalg.tensorsolve(AR_i8, AR_i8), npt.NDArray[np.float64])
assert_type(np.linalg.tensorsolve(AR_i8, AR_f8), npt.NDArray[np.float64])
assert_type(np.linalg.tensorsolve(AR_f4, AR_f4), npt.NDArray[np.float32])
assert_type(np.linalg.tensorsolve(AR_c16, AR_f8), npt.NDArray[np.complex128])
assert_type(np.linalg.tensorsolve(AR_c8, AR_f4), npt.NDArray[np.complex64])
assert_type(np.linalg.tensorsolve(AR_f4, AR_c8), npt.NDArray[np.complex64])

assert_type(np.linalg.solve(AR_i8, AR_i8), npt.NDArray[np.float64])
assert_type(np.linalg.solve(AR_i8, AR_f8), npt.NDArray[np.float64])
assert_type(np.linalg.solve(AR_f4, AR_f4), npt.NDArray[np.float32])
assert_type(np.linalg.solve(AR_c16, AR_f8), npt.NDArray[np.complex128])
assert_type(np.linalg.solve(AR_c8, AR_f4), npt.NDArray[np.complex64])
assert_type(np.linalg.solve(AR_f4, AR_c8), npt.NDArray[np.complex64])

assert_type(np.linalg.tensorinv(AR_i8), npt.NDArray[np.float64])
assert_type(np.linalg.tensorinv(AR_f8), npt.NDArray[np.float64])
assert_type(np.linalg.tensorinv(AR_c16), npt.NDArray[np.complex128])

assert_type(np.linalg.inv(AR_i8), npt.NDArray[np.float64])
assert_type(np.linalg.inv(AR_f8), npt.NDArray[np.float64])
assert_type(np.linalg.inv(AR_c16), npt.NDArray[np.complex128])

assert_type(np.linalg.pinv(AR_i8), npt.NDArray[np.float64])
assert_type(np.linalg.pinv(AR_f8), npt.NDArray[np.float64])
assert_type(np.linalg.pinv(AR_c16), npt.NDArray[np.complex128])

assert_type(np.linalg.matrix_power(AR_i8, -1), npt.NDArray[np.float64])
assert_type(np.linalg.matrix_power(AR_i8, 1), npt.NDArray[np.int64])
assert_type(np.linalg.matrix_power(AR_f8, 0), npt.NDArray[np.float64])
assert_type(np.linalg.matrix_power(AR_c16, 1), npt.NDArray[np.complex128])
assert_type(np.linalg.matrix_power(AR_O, 2), npt.NDArray[np.object_])

assert_type(np.linalg.cholesky(AR_i8), npt.NDArray[np.float64])
assert_type(np.linalg.cholesky(AR_f8), npt.NDArray[np.float64])
assert_type(np.linalg.cholesky(AR_c16), npt.NDArray[np.complex128])

assert_type(np.linalg.qr(AR_i8), QRResult[np.float64])
assert_type(np.linalg.qr(AR_i8, "r"), npt.NDArray[np.float64])
assert_type(np.linalg.qr(AR_i8, "raw"), tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]])
assert_type(np.linalg.qr(AR_f4), QRResult[np.float32])
assert_type(np.linalg.qr(AR_f4, "r"), npt.NDArray[np.float32])
assert_type(np.linalg.qr(AR_f4, "raw"), tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]])
assert_type(np.linalg.qr(AR_f8), QRResult[np.float64])
assert_type(np.linalg.qr(AR_f8, "r"), npt.NDArray[np.float64])
assert_type(np.linalg.qr(AR_f8, "raw"), tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]])
assert_type(np.linalg.qr(AR_c8), QRResult[np.complex64])
assert_type(np.linalg.qr(AR_c8, "r"), npt.NDArray[np.complex64])
assert_type(np.linalg.qr(AR_c8, "raw"), tuple[npt.NDArray[np.complex64], npt.NDArray[np.complex64]])
assert_type(np.linalg.qr(AR_c16), QRResult[np.complex128])
assert_type(np.linalg.qr(AR_c16, "r"), npt.NDArray[np.complex128])
assert_type(np.linalg.qr(AR_c16, "raw"), tuple[npt.NDArray[np.complex128], npt.NDArray[np.complex128]])
# Mypy bug: `Expression is of type "QRResult[Any]", not "QRResult[Any]"`
assert_type(np.linalg.qr(AR_any), QRResult[Any])  # type: ignore[assert-type]
# Mypy bug: `Expression is of type "ndarray[Any, Any]", not "ndarray[tuple[Any, ...], dtype[Any]]"`
assert_type(np.linalg.qr(AR_any, "r"), npt.NDArray[Any])  # type: ignore[assert-type]
# Mypy bug: `Expression is of type "tuple[Any, ...]", <--snip-->"`
assert_type(np.linalg.qr(AR_any, "raw"), tuple[npt.NDArray[Any], npt.NDArray[Any]])  # type: ignore[assert-type]

assert_type(np.linalg.eigvals(AR_i8), npt.NDArray[np.float64] | npt.NDArray[np.complex128])
assert_type(np.linalg.eigvals(AR_f8), npt.NDArray[np.float64] | npt.NDArray[np.complex128])
assert_type(np.linalg.eigvals(AR_c16), npt.NDArray[np.complex128])

assert_type(np.linalg.eigvalsh(AR_i8), npt.NDArray[np.float64])
assert_type(np.linalg.eigvalsh(AR_f8), npt.NDArray[np.float64])
assert_type(np.linalg.eigvalsh(AR_c16), npt.NDArray[np.float64])

assert_type(np.linalg.eig(AR_i8), EigResult[np.complex128])
assert_type(np.linalg.eig(AR_f4), EigResult[np.complex64])
assert_type(np.linalg.eig(AR_f8), EigResult[np.complex128])
assert_type(np.linalg.eig(AR_c8), EigResult[np.complex64])
assert_type(np.linalg.eig(AR_c16), EigResult[np.complex128])
# Mypy bug: `Expression is of type "EigResult[Any]", not "EigResult[Any]"`
assert_type(np.linalg.eig(AR_f_), EigResult[Any])  # type: ignore[assert-type]
assert_type(np.linalg.eig(AR_c_), EigResult[Any])  # type: ignore[assert-type]
assert_type(np.linalg.eig(AR_any), EigResult[Any])  # type: ignore[assert-type]

assert_type(np.linalg.eigh(AR_i8), EighResult[np.float64, np.float64])
assert_type(np.linalg.eigh(AR_f4), EighResult[np.float32, np.float32])
assert_type(np.linalg.eigh(AR_f8), EighResult[np.float64, np.float64])
assert_type(np.linalg.eigh(AR_c8), EighResult[np.float32, np.complex64])
assert_type(np.linalg.eigh(AR_c16), EighResult[np.float64, np.complex128])
# Mypy bug: `Expression is of type "EighResult[Any, Any]", not "EighResult[Any, Any]"`
assert_type(np.linalg.eigh(AR_any), EighResult[Any, Any])  # type: ignore[assert-type]

assert_type(np.linalg.svd(AR_i8), SVDResult[np.float64, np.float64])
assert_type(np.linalg.svd(AR_i8, compute_uv=False), npt.NDArray[np.float64])
assert_type(np.linalg.svd(AR_f4), SVDResult[np.float32, np.float32])
assert_type(np.linalg.svd(AR_f4, compute_uv=False), npt.NDArray[np.float32])
assert_type(np.linalg.svd(AR_f8), SVDResult[np.float64, np.float64])
assert_type(np.linalg.svd(AR_f8, compute_uv=False), npt.NDArray[np.float64])
assert_type(np.linalg.svd(AR_c8), SVDResult[np.float32, np.complex64])
assert_type(np.linalg.svd(AR_c8, compute_uv=False), npt.NDArray[np.float32])
assert_type(np.linalg.svd(AR_c16), SVDResult[np.float64, np.complex128])
assert_type(np.linalg.svd(AR_c16, compute_uv=False), npt.NDArray[np.float64])
assert_type(np.linalg.svd(int_list_2d), SVDResult[np.float64, np.float64])
assert_type(np.linalg.svd(int_list_2d, compute_uv=False), npt.NDArray[np.float64])
assert_type(np.linalg.svd(float_list_2d), SVDResult[np.float64, np.float64])
assert_type(np.linalg.svd(float_list_2d, compute_uv=False), npt.NDArray[np.float64])
assert_type(np.linalg.svd(complex_list_2d), SVDResult[np.float64, np.complex128])
assert_type(np.linalg.svd(complex_list_2d, compute_uv=False), npt.NDArray[np.float64])
# Mypy bug: `Expression is of type "SVDResult[Any, Any]", not "SVDResult[Any, Any]"`
assert_type(np.linalg.svd(AR_any), SVDResult[Any, Any])  # type: ignore[assert-type]
# Mypy bug: `Expression is of type "ndarray[Any, Any]", not "ndarray[tuple[Any, ...], dtype[Any]]"`
assert_type(np.linalg.svd(AR_any, compute_uv=False), npt.NDArray[Any])  # type: ignore[assert-type]

assert_type(np.linalg.svdvals(AR_b), npt.NDArray[np.float64])
assert_type(np.linalg.svdvals(AR_i8), npt.NDArray[np.float64])
assert_type(np.linalg.svdvals(AR_f4), npt.NDArray[np.float32])
assert_type(np.linalg.svdvals(AR_f8), npt.NDArray[np.float64])
assert_type(np.linalg.svdvals(AR_c8), npt.NDArray[np.float32])
assert_type(np.linalg.svdvals(AR_c16), npt.NDArray[np.float64])
assert_type(np.linalg.svdvals(int_list_2d), npt.NDArray[np.float64])
assert_type(np.linalg.svdvals(float_list_2d), npt.NDArray[np.float64])
assert_type(np.linalg.svdvals(complex_list_2d), npt.NDArray[np.float64])

assert_type(np.linalg.matrix_rank(AR_i8), Any)
assert_type(np.linalg.matrix_rank(AR_f8), Any)
assert_type(np.linalg.matrix_rank(AR_c16), Any)
assert_type(np.linalg.matrix_rank(SC_f8), Literal[0, 1])
assert_type(np.linalg.matrix_rank(AR_f8_1d), Literal[0, 1])
assert_type(np.linalg.matrix_rank(float_list_1d), Literal[0, 1])
assert_type(np.linalg.matrix_rank(AR_f8_2d), np.int_)
assert_type(np.linalg.matrix_rank(float_list_2d), np.int_)
assert_type(np.linalg.matrix_rank(AR_f8_3d), _Array1D[np.int_])
assert_type(np.linalg.matrix_rank(float_list_3d), _Array1D[np.int_])
assert_type(np.linalg.matrix_rank(AR_f8_4d), npt.NDArray[np.int_])
assert_type(np.linalg.matrix_rank(float_list_4d), npt.NDArray[np.int_])

assert_type(np.linalg.cond(AR_i8), Any)
assert_type(np.linalg.cond(AR_f8), Any)
assert_type(np.linalg.cond(AR_c16), Any)
assert_type(np.linalg.cond(AR_f4_2d), np.float32)
assert_type(np.linalg.cond(AR_f8_2d), np.float64)
assert_type(np.linalg.cond(AR_f4_3d), npt.NDArray[np.float32])
assert_type(np.linalg.cond(AR_f8_3d), npt.NDArray[np.float64])

assert_type(np.linalg.slogdet(AR_i8), SlogdetResult)
assert_type(np.linalg.slogdet(AR_f8), SlogdetResult)
assert_type(np.linalg.slogdet(AR_c16), SlogdetResult)
assert_type(np.linalg.slogdet(AR_f4_2d), SlogdetResult[np.float32, np.float32])
assert_type(np.linalg.slogdet(AR_f8_2d), SlogdetResult[np.float64, np.float64])
assert_type(np.linalg.slogdet(AR_f4_3d), SlogdetResult[npt.NDArray[np.float32], npt.NDArray[np.float32]])
assert_type(np.linalg.slogdet(AR_f8_3d), SlogdetResult[npt.NDArray[np.float64], npt.NDArray[np.float64]])
assert_type(np.linalg.slogdet(complex_list_2d), SlogdetResult[np.float64, np.complex128])
assert_type(np.linalg.slogdet(complex_list_3d), SlogdetResult[npt.NDArray[np.float64], npt.NDArray[np.complex128]])

assert_type(np.linalg.det(AR_i8), Any)
assert_type(np.linalg.det(AR_f8), Any)
assert_type(np.linalg.det(AR_c16), Any)
assert_type(np.linalg.det(AR_f4_2d), np.float32)
assert_type(np.linalg.det(AR_f8_2d), np.float64)
assert_type(np.linalg.det(AR_f4_3d), npt.NDArray[np.float32])
assert_type(np.linalg.det(AR_f8_3d), npt.NDArray[np.float64])
assert_type(np.linalg.det(complex_list_2d), np.complex128)
assert_type(np.linalg.det(complex_list_3d), npt.NDArray[np.complex128])

assert_type(
    np.linalg.lstsq(AR_i8, AR_i8),
    tuple[npt.NDArray[np.float64], _Array1D[np.float64], np.int32, _Array1D[np.float64]],
)
assert_type(
    np.linalg.lstsq(AR_f4, AR_f4),
    tuple[npt.NDArray[np.float32], _Array1D[np.float32], np.int32, _Array1D[np.float32]],
)
assert_type(
    np.linalg.lstsq(AR_i8, AR_f8),
    tuple[npt.NDArray[np.float64], _Array1D[np.float64], np.int32, _Array1D[np.float64]],
)
assert_type(
    np.linalg.lstsq(AR_f4, AR_f8),
    tuple[npt.NDArray[np.float64], _Array1D[np.float64], np.int32, _Array1D[np.float64]],
)
assert_type(
    np.linalg.lstsq(AR_f8, AR_i8),
    tuple[npt.NDArray[np.float64], _Array1D[np.float64], np.int32, _Array1D[np.float64]],
)
assert_type(
    np.linalg.lstsq(AR_f8, AR_f4),
    tuple[npt.NDArray[np.float64], _Array1D[np.float64], np.int32, _Array1D[np.float64]],
)
assert_type(
    np.linalg.lstsq(AR_c8, AR_c8),
    tuple[npt.NDArray[np.complex64], _Array1D[np.float32], np.int32, _Array1D[np.float32]],
)
assert_type(
    np.linalg.lstsq(AR_c8, AR_c16),
    tuple[npt.NDArray[np.complex128], _Array1D[np.float64], np.int32, _Array1D[np.float64]],
)
assert_type(
    np.linalg.lstsq(AR_c16, AR_c8),
    tuple[npt.NDArray[np.complex128], _Array1D[np.float64], np.int32, _Array1D[np.float64]],
)
assert_type(
    np.linalg.lstsq(AR_f8, AR_f8_1d),
    tuple[_Array1D[np.float64], _Array1D[np.float64], np.int32, _Array1D[np.float64]],
)
assert_type(
    np.linalg.lstsq(AR_f4, AR_f4_1d),
    tuple[_Array1D[np.float32], _Array1D[np.float32], np.int32, _Array1D[np.float32]],
)
assert_type(
    np.linalg.lstsq(AR_f8, AR_f8_2d),
    tuple[_Array2D[np.float64], _Array1D[np.float64], np.int32, _Array1D[np.float64]],
)
assert_type(
    np.linalg.lstsq(AR_f4, AR_f4_2d),
    tuple[_Array2D[np.float32], _Array1D[np.float32], np.int32, _Array1D[np.float32]],
)

assert_type(np.linalg.norm(AR_i8), np.float64)
assert_type(np.linalg.norm(AR_f8), np.float64)
assert_type(np.linalg.norm(AR_c16), np.float64)
# Mypy incorrectly infers `Any` for datetime64 and timedelta64, but pyright behaves correctly.
assert_type(np.linalg.norm(AR_M), np.float64)  # type: ignore[assert-type]
assert_type(np.linalg.norm(AR_m), np.float64)  # type: ignore[assert-type]
assert_type(np.linalg.norm(AR_U), np.float64)
assert_type(np.linalg.norm(AR_S), np.float64)
assert_type(np.linalg.norm(AR_f8, 0, 1), npt.NDArray[np.float64])
assert_type(np.linalg.norm(AR_f8, axis=0), npt.NDArray[np.float64])
assert_type(np.linalg.norm(AR_f8, keepdims=True), npt.NDArray[np.float64])
assert_type(np.linalg.norm(AR_f8_2d, keepdims=True), _Array2D[np.float64])
assert_type(np.linalg.norm(AR_f2), np.float16)
assert_type(np.linalg.norm(AR_f2, 0, 1), npt.NDArray[np.float16])
assert_type(np.linalg.norm(AR_f2, axis=1), npt.NDArray[np.float16])
assert_type(np.linalg.norm(AR_f2, keepdims=True), npt.NDArray[np.float16])
assert_type(np.linalg.norm(AR_f2_2d, keepdims=True), _Array2D[np.float16])
assert_type(np.linalg.norm(AR_f4), np.float32)
assert_type(np.linalg.norm(AR_c8), np.float32)
assert_type(np.linalg.norm(AR_f4, 0, 1), npt.NDArray[np.float32])
assert_type(np.linalg.norm(AR_f4, axis=1), npt.NDArray[np.float32])
assert_type(np.linalg.norm(AR_f4, keepdims=True), npt.NDArray[np.float32])
assert_type(np.linalg.norm(AR_f4_2d, keepdims=True), _Array2D[np.float32])
assert_type(np.linalg.norm(AR_f10), np.longdouble)
assert_type(np.linalg.norm(AR_c20), np.longdouble)
assert_type(np.linalg.norm(AR_f10, 0, 1), npt.NDArray[np.longdouble])
assert_type(np.linalg.norm(AR_f10, axis=1), npt.NDArray[np.longdouble])
assert_type(np.linalg.norm(AR_f10, keepdims=True), npt.NDArray[np.longdouble])
assert_type(np.linalg.norm(AR_f10_2d, keepdims=True), _Array2D[np.longdouble])

assert_type(np.linalg.matrix_norm(AR_i8), npt.NDArray[np.float64] | Any)
assert_type(np.linalg.matrix_norm(AR_f8), npt.NDArray[np.float64] | Any)
assert_type(np.linalg.matrix_norm(AR_c16), npt.NDArray[np.float64] | Any)
assert_type(np.linalg.matrix_norm(AR_U), npt.NDArray[np.float64] | Any)
assert_type(np.linalg.matrix_norm(AR_S), npt.NDArray[np.float64] | Any)
assert_type(np.linalg.matrix_norm(AR_f8_2d), np.float64)
assert_type(np.linalg.matrix_norm(AR_f8_3d), npt.NDArray[np.float64])
assert_type(np.linalg.matrix_norm(AR_f8_2d, keepdims=True), _Array2D[np.float64])
assert_type(np.linalg.matrix_norm(AR_f4), npt.NDArray[np.float32] | Any)
assert_type(np.linalg.matrix_norm(AR_c8), npt.NDArray[np.float32] | Any)
assert_type(np.linalg.matrix_norm(AR_f4_2d), np.float32)
assert_type(np.linalg.matrix_norm(AR_f4_3d), npt.NDArray[np.float32])
assert_type(np.linalg.matrix_norm(AR_f4_2d, keepdims=True), _Array2D[np.float32])
assert_type(np.linalg.matrix_norm(AR_f10), npt.NDArray[np.longdouble] | Any)
assert_type(np.linalg.matrix_norm(AR_c20), npt.NDArray[np.longdouble] | Any)
assert_type(np.linalg.matrix_norm(AR_f10_2d), np.longdouble)
assert_type(np.linalg.matrix_norm(AR_f10_3d), npt.NDArray[np.longdouble])
assert_type(np.linalg.matrix_norm(AR_f10_2d, keepdims=True), _Array2D[np.longdouble])
assert_type(np.linalg.matrix_norm(complex_list_2d), np.float64)
assert_type(np.linalg.matrix_norm(complex_list_3d), npt.NDArray[np.float64])
assert_type(np.linalg.matrix_norm(complex_list_2d, keepdims=True), npt.NDArray[np.float64])

assert_type(np.linalg.vector_norm(AR_i8), np.float64)
assert_type(np.linalg.vector_norm(AR_f8), np.float64)
assert_type(np.linalg.vector_norm(AR_c16), np.float64)
# Mypy incorrectly infers `Any` for datetime64 and timedelta64, but pyright behaves correctly.
assert_type(np.linalg.vector_norm(AR_M), np.float64)  # type: ignore[assert-type]
assert_type(np.linalg.vector_norm(AR_m), np.float64)  # type: ignore[assert-type]
assert_type(np.linalg.vector_norm(AR_U), np.float64)
assert_type(np.linalg.vector_norm(AR_S), np.float64)
assert_type(np.linalg.vector_norm(AR_f8, axis=0), npt.NDArray[np.float64])
assert_type(np.linalg.vector_norm(AR_f8, keepdims=True), npt.NDArray[np.float64])
assert_type(np.linalg.vector_norm(AR_f8_2d, keepdims=True), _Array2D[np.float64])
assert_type(np.linalg.vector_norm(AR_f2), np.float16)
assert_type(np.linalg.vector_norm(AR_f2, axis=1), npt.NDArray[np.float16])
assert_type(np.linalg.vector_norm(AR_f2, keepdims=True), npt.NDArray[np.float16])
assert_type(np.linalg.vector_norm(AR_f2_2d, keepdims=True), _Array2D[np.float16])
assert_type(np.linalg.vector_norm(AR_f4), np.float32)
assert_type(np.linalg.vector_norm(AR_c8), np.float32)
assert_type(np.linalg.vector_norm(AR_f4, axis=1), npt.NDArray[np.float32])
assert_type(np.linalg.vector_norm(AR_f4, keepdims=True), npt.NDArray[np.float32])
assert_type(np.linalg.vector_norm(AR_f4_2d, keepdims=True), _Array2D[np.float32])
assert_type(np.linalg.vector_norm(AR_f10), np.longdouble)
assert_type(np.linalg.vector_norm(AR_c20), np.longdouble)
assert_type(np.linalg.vector_norm(AR_f10, axis=1), npt.NDArray[np.longdouble])
assert_type(np.linalg.vector_norm(AR_f10, keepdims=True), npt.NDArray[np.longdouble])
assert_type(np.linalg.vector_norm(AR_f10_2d, keepdims=True), _Array2D[np.longdouble])

assert_type(np.linalg.tensordot(AR_b, AR_b), npt.NDArray[np.bool])
assert_type(np.linalg.tensordot(AR_i8, AR_i8), npt.NDArray[np.int64])
assert_type(np.linalg.tensordot(AR_f8, AR_f8), npt.NDArray[np.float64])
assert_type(np.linalg.tensordot(AR_c16, AR_c16), npt.NDArray[np.complex128])
assert_type(np.linalg.tensordot(AR_m, AR_m), npt.NDArray[np.timedelta64])
assert_type(np.linalg.tensordot(AR_O, AR_O), npt.NDArray[np.object_])

assert_type(np.linalg.multi_dot([AR_i8, AR_i8]), npt.NDArray[np.int64])
assert_type(np.linalg.multi_dot([AR_f8, AR_f8]), npt.NDArray[np.float64])
assert_type(np.linalg.multi_dot([AR_c16, AR_c16]), npt.NDArray[np.complex128])
assert_type(np.linalg.multi_dot([AR_O, AR_O]), npt.NDArray[np.object_])
# Mypy incorrectly infers `ndarray[Any, Any]`, but pyright behaves correctly.
assert_type(np.linalg.multi_dot([AR_i8, AR_f8]), npt.NDArray[np.float64 | Any])  # type: ignore[assert-type]
assert_type(np.linalg.multi_dot([AR_f8, AR_c16]), npt.NDArray[np.complex128 | Any])  # type: ignore[assert-type]
assert_type(np.linalg.multi_dot([AR_m, AR_m]), npt.NDArray[np.timedelta64])  # type: ignore[assert-type]

# Mypy incorrectly infers `ndarray[Any, Any]`, but pyright behaves correctly.
assert_type(np.linalg.diagonal(AR_any), np.ndarray)  # type: ignore[assert-type]
assert_type(np.linalg.diagonal(AR_f4), npt.NDArray[np.float32])
assert_type(np.linalg.diagonal(AR_f4_2d), _Array1D[np.float32])
assert_type(np.linalg.diagonal(AR_f8_2d), _Array1D[np.float64])
assert_type(np.linalg.diagonal(bool_list_2d), npt.NDArray[np.bool])
assert_type(np.linalg.diagonal(int_list_2d), npt.NDArray[np.int_])
assert_type(np.linalg.diagonal(float_list_2d), npt.NDArray[np.float64])
assert_type(np.linalg.diagonal(complex_list_2d), npt.NDArray[np.complex128])
assert_type(np.linalg.diagonal(bytes_list_2d), npt.NDArray[np.bytes_])
assert_type(np.linalg.diagonal(str_list_2d), npt.NDArray[np.str_])

assert_type(np.linalg.trace(AR_any), Any)
assert_type(np.linalg.trace(AR_f4), Any)
assert_type(np.linalg.trace(AR_f4_2d), np.float32)
assert_type(np.linalg.trace(AR_f8_2d), np.float64)
assert_type(np.linalg.trace(AR_f4_3d), _Array1D[np.float32])
assert_type(np.linalg.trace(AR_f8_3d), _Array1D[np.float64])
assert_type(np.linalg.trace(AR_f8_4d), np.ndarray[tuple[int, *tuple[Any, ...]], np.dtype[np.float64]])
assert_type(np.linalg.trace(bool_list_2d), np.bool)
assert_type(np.linalg.trace(int_list_2d), np.int_)
assert_type(np.linalg.trace(float_list_2d), np.float64)
assert_type(np.linalg.trace(complex_list_2d), np.complex128)
assert_type(np.linalg.trace(float_list_3d), npt.NDArray[np.float64])

assert_type(np.linalg.outer(bool_list_1d, bool_list_1d), _Array2D[np.bool])
assert_type(np.linalg.outer(int_list_1d, int_list_1d), _Array2D[np.int64])
assert_type(np.linalg.outer(float_list_1d, float_list_1d), _Array2D[np.float64])
assert_type(np.linalg.outer(complex_list_1d, complex_list_1d), _Array2D[np.complex128])
assert_type(np.linalg.outer(AR_i8, AR_i8), _Array2D[np.int64])
assert_type(np.linalg.outer(AR_f8, AR_f8), _Array2D[np.float64])
assert_type(np.linalg.outer(AR_c16, AR_c16), _Array2D[np.complex128])
assert_type(np.linalg.outer(AR_b, AR_b), _Array2D[np.bool])
assert_type(np.linalg.outer(AR_O, AR_O), _Array2D[np.object_])
assert_type(np.linalg.outer(AR_i8, AR_m), _Array2D[np.timedelta64])

assert_type(np.linalg.cross(int_list_1d, int_list_1d), npt.NDArray[np.int64])
assert_type(np.linalg.cross(float_list_1d, int_list_1d), npt.NDArray[np.float64])
assert_type(np.linalg.cross(float_list_1d, complex_list_1d), npt.NDArray[np.complex128])
assert_type(np.linalg.cross(AR_i8, AR_i8), npt.NDArray[np.int64])
assert_type(np.linalg.cross(AR_f8, AR_i8), npt.NDArray[np.float64])
assert_type(np.linalg.cross(AR_f2, AR_f2), npt.NDArray[np.float16])
assert_type(np.linalg.cross(AR_f8, AR_f8), npt.NDArray[np.float64])
assert_type(np.linalg.cross(AR_c16, AR_i8), npt.NDArray[np.complex128])
assert_type(np.linalg.cross(AR_c16, AR_f8), npt.NDArray[np.complex128])
assert_type(np.linalg.cross(AR_c16, AR_c16), npt.NDArray[np.complex128])
assert_type(np.linalg.cross(AR_O, AR_f8), npt.NDArray[np.object_])
assert_type(np.linalg.cross(AR_f8, AR_O), npt.NDArray[np.object_])
assert_type(np.linalg.cross(AR_O, AR_O), npt.NDArray[np.object_])

assert_type(np.linalg.matmul(AR_b, AR_b), npt.NDArray[np.bool] | Any)
assert_type(np.linalg.matmul(AR_i8, AR_b), npt.NDArray[np.int64] | Any)
assert_type(np.linalg.matmul(AR_b, AR_i8), npt.NDArray[np.int64] | Any)
assert_type(np.linalg.matmul(AR_i8, AR_i8), npt.NDArray[np.int64] | Any)
assert_type(np.linalg.matmul(AR_f4, AR_f4), npt.NDArray[np.float32] | Any)
assert_type(np.linalg.matmul(AR_f8, AR_i8), npt.NDArray[np.float64] | Any)
assert_type(np.linalg.matmul(AR_i8, AR_f8), npt.NDArray[np.float64] | Any)
assert_type(np.linalg.matmul(AR_f8, AR_f8), npt.NDArray[np.float64] | Any)
assert_type(np.linalg.matmul(AR_c16, AR_i8), npt.NDArray[np.complex128] | Any)
assert_type(np.linalg.matmul(AR_f8, AR_c16), npt.NDArray[np.complex128] | Any)
assert_type(np.linalg.matmul(AR_c16, AR_c16), npt.NDArray[np.complex128] | Any)
# same as the block above, but for the 1d x 1d case
assert_type(np.linalg.matmul(AR_b_1d, AR_b_1d), np.bool)
assert_type(np.linalg.matmul(AR_i8_1d, AR_b_1d), np.int64)
assert_type(np.linalg.matmul(AR_b_1d, AR_i8_1d), np.int64)
assert_type(np.linalg.matmul(AR_i8_1d, AR_i8_1d), np.int64)
assert_type(np.linalg.matmul(AR_f4_1d, AR_f4_1d), np.float32)
assert_type(np.linalg.matmul(AR_f8_1d, AR_i8_1d), np.float64)
assert_type(np.linalg.matmul(AR_i8_1d, AR_f8_1d), np.float64)
assert_type(np.linalg.matmul(AR_f8_1d, AR_f8_1d), np.float64)
assert_type(np.linalg.matmul(AR_c16_1d, AR_i8_1d), np.complex128)
assert_type(np.linalg.matmul(AR_f8_1d, AR_c16_1d), np.complex128)
assert_type(np.linalg.matmul(AR_c16_1d, AR_c16_1d), np.complex128)
# 1d x 2d
assert_type(np.linalg.matmul(AR_b_1d, AR_b_2d), npt.NDArray[np.bool])
assert_type(np.linalg.matmul(AR_i8_1d, AR_b_2d), npt.NDArray[np.int64])
assert_type(np.linalg.matmul(AR_b_1d, AR_i8_2d), npt.NDArray[np.int64])
assert_type(np.linalg.matmul(AR_i8_1d, AR_i8_2d), npt.NDArray[np.int64])
assert_type(np.linalg.matmul(AR_f4_1d, AR_f4_2d), npt.NDArray[np.float32])
assert_type(np.linalg.matmul(AR_f8_1d, AR_i8_2d), npt.NDArray[np.float64])
assert_type(np.linalg.matmul(AR_i8_1d, AR_f8_2d), npt.NDArray[np.float64])
assert_type(np.linalg.matmul(AR_f8_1d, AR_f8_2d), npt.NDArray[np.float64])
assert_type(np.linalg.matmul(AR_c16_1d, AR_i8_2d), npt.NDArray[np.complex128])
assert_type(np.linalg.matmul(AR_f8_1d, AR_c16_2d), npt.NDArray[np.complex128])
assert_type(np.linalg.matmul(AR_c16_1d, AR_c16_2d), npt.NDArray[np.complex128])
# 1d x ?d
assert_type(np.linalg.matmul(AR_b_1d, AR_b), npt.NDArray[np.bool] | Any)
assert_type(np.linalg.matmul(AR_i8_1d, AR_b), npt.NDArray[np.int64] | Any)
assert_type(np.linalg.matmul(AR_b_1d, AR_i8), npt.NDArray[np.int64] | Any)
assert_type(np.linalg.matmul(AR_i8_1d, AR_i8), npt.NDArray[np.int64] | Any)
assert_type(np.linalg.matmul(AR_f4_1d, AR_f4), npt.NDArray[np.float32] | Any)
assert_type(np.linalg.matmul(AR_f8_1d, AR_i8), npt.NDArray[np.float64] | Any)
assert_type(np.linalg.matmul(AR_i8_1d, AR_f8), npt.NDArray[np.float64] | Any)
assert_type(np.linalg.matmul(AR_f8_1d, AR_f8), npt.NDArray[np.float64] | Any)
assert_type(np.linalg.matmul(AR_c16_1d, AR_i8), npt.NDArray[np.complex128] | Any)
assert_type(np.linalg.matmul(AR_f8_1d, AR_c16), npt.NDArray[np.complex128] | Any)
assert_type(np.linalg.matmul(AR_c16_1d, AR_c16), npt.NDArray[np.complex128] | Any)
# 2d x 1d
assert_type(np.linalg.matmul(AR_b_2d, AR_b_1d), npt.NDArray[np.bool])
assert_type(np.linalg.matmul(AR_i8_2d, AR_b_1d), npt.NDArray[np.int64])
assert_type(np.linalg.matmul(AR_b_2d, AR_i8_1d), npt.NDArray[np.int64])
assert_type(np.linalg.matmul(AR_i8_2d, AR_i8_1d), npt.NDArray[np.int64])
assert_type(np.linalg.matmul(AR_f4_2d, AR_f4_1d), npt.NDArray[np.float32])
assert_type(np.linalg.matmul(AR_f8_2d, AR_i8_1d), npt.NDArray[np.float64])
assert_type(np.linalg.matmul(AR_i8_2d, AR_f8_1d), npt.NDArray[np.float64])
assert_type(np.linalg.matmul(AR_f8_2d, AR_f8_1d), npt.NDArray[np.float64])
assert_type(np.linalg.matmul(AR_c16_2d, AR_i8_1d), npt.NDArray[np.complex128])
assert_type(np.linalg.matmul(AR_f8_2d, AR_c16_1d), npt.NDArray[np.complex128])
assert_type(np.linalg.matmul(AR_c16_2d, AR_c16_1d), npt.NDArray[np.complex128])
# 2d x ?d
assert_type(np.linalg.matmul(AR_b_2d, AR_b), npt.NDArray[np.bool])
assert_type(np.linalg.matmul(AR_i8_2d, AR_b), npt.NDArray[np.int64])
assert_type(np.linalg.matmul(AR_b_2d, AR_i8), npt.NDArray[np.int64])
assert_type(np.linalg.matmul(AR_i8_2d, AR_i8), npt.NDArray[np.int64])
assert_type(np.linalg.matmul(AR_f4_2d, AR_f4), npt.NDArray[np.float32])
assert_type(np.linalg.matmul(AR_f8_2d, AR_i8), npt.NDArray[np.float64])
assert_type(np.linalg.matmul(AR_i8_2d, AR_f8), npt.NDArray[np.float64])
assert_type(np.linalg.matmul(AR_f8_2d, AR_f8), npt.NDArray[np.float64])
assert_type(np.linalg.matmul(AR_c16_2d, AR_i8), npt.NDArray[np.complex128])
assert_type(np.linalg.matmul(AR_f8_2d, AR_c16), npt.NDArray[np.complex128])
assert_type(np.linalg.matmul(AR_c16_2d, AR_c16), npt.NDArray[np.complex128])
# ?d x 1d
assert_type(np.linalg.matmul(AR_b, AR_b_1d), npt.NDArray[np.bool] | Any)
assert_type(np.linalg.matmul(AR_i8, AR_b_1d), npt.NDArray[np.int64] | Any)
assert_type(np.linalg.matmul(AR_b, AR_i8_1d), npt.NDArray[np.int64] | Any)
assert_type(np.linalg.matmul(AR_i8, AR_i8_1d), npt.NDArray[np.int64] | Any)
assert_type(np.linalg.matmul(AR_f4, AR_f4_1d), npt.NDArray[np.float32] | Any)
assert_type(np.linalg.matmul(AR_f8, AR_i8_1d), npt.NDArray[np.float64] | Any)
assert_type(np.linalg.matmul(AR_i8, AR_f8_1d), npt.NDArray[np.float64] | Any)
assert_type(np.linalg.matmul(AR_f8, AR_f8_1d), npt.NDArray[np.float64] | Any)
assert_type(np.linalg.matmul(AR_c16, AR_i8_1d), npt.NDArray[np.complex128] | Any)
assert_type(np.linalg.matmul(AR_f8, AR_c16_1d), npt.NDArray[np.complex128] | Any)
assert_type(np.linalg.matmul(AR_c16, AR_c16_1d), npt.NDArray[np.complex128] | Any)



from typing import Any, ClassVar, Final, overload

import numpy as np
import numpy.typing as npt
from numpy._typing import (
    _ArrayLikeFloat_co,
    _ArrayLikeNumber_co,
    _FloatLike_co,
    _NumberLike_co,
)

from ._polybase import ABCPolyBase
from ._polytypes import (
    _Array1,
    _Array2,
    _ArrayLikeCoef_co,
    _FuncBinOp,
    _FuncCompanion,
    _FuncDer,
    _FuncFit,
    _FuncFromRoots,
    _FuncInteg,
    _FuncLine,
    _FuncPow,
    _FuncRoots,
    _FuncUnOp,
    _FuncVal,
    _FuncVal2D,
    _FuncVal3D,
    _FuncValND,
    _FuncVander,
    _FuncVander2D,
    _FuncVander3D,
)
from .polyutils import trimcoef as polytrim

__all__ = [
    "polyzero",
    "polyone",
    "polyx",
    "polydomain",
    "polyline",
    "polyadd",
    "polysub",
    "polymulx",
    "polymul",
    "polydiv",
    "polypow",
    "polyval",
    "polyvalfromroots",
    "polyder",
    "polyint",
    "polyfromroots",
    "polyvander",
    "polyfit",
    "polytrim",
    "polyroots",
    "Polynomial",
    "polyval2d",
    "polyval3d",
    "polyvalnd",
    "polygrid2d",
    "polygrid3d",
    "polyvander2d",
    "polyvander3d",
    "polycompanion",
]

polydomain: Final[_Array2[np.float64]] = ...
polyzero: Final[_Array1[np.int_]] = ...
polyone: Final[_Array1[np.int_]] = ...
polyx: Final[_Array2[np.int_]] = ...

polyline: Final[_FuncLine] = ...
polyfromroots: Final[_FuncFromRoots] = ...
polyadd: Final[_FuncBinOp] = ...
polysub: Final[_FuncBinOp] = ...
polymulx: Final[_FuncUnOp] = ...
polymul: Final[_FuncBinOp] = ...
polydiv: Final[_FuncBinOp] = ...
polypow: Final[_FuncPow] = ...
polyder: Final[_FuncDer] = ...
polyint: Final[_FuncInteg] = ...
polyval: Final[_FuncVal] = ...
polyval2d: Final[_FuncVal2D] = ...
polyval3d: Final[_FuncVal3D] = ...
polyvalnd: Final[_FuncValND] = ...

@overload
def polyvalfromroots(x: _FloatLike_co, r: _FloatLike_co, tensor: bool = True) -> np.float64 | Any: ...
@overload
def polyvalfromroots(x: _NumberLike_co, r: _NumberLike_co, tensor: bool = True) -> np.complex128 | Any: ...
@overload
def polyvalfromroots(x: _ArrayLikeFloat_co, r: _ArrayLikeFloat_co, tensor: bool = True) -> npt.NDArray[np.float64 | Any]: ...
@overload
def polyvalfromroots(x: _ArrayLikeNumber_co, r: _ArrayLikeNumber_co, tensor: bool = True) -> npt.NDArray[np.complex128 | Any]: ...
@overload
def polyvalfromroots(x: _ArrayLikeCoef_co, r: _ArrayLikeCoef_co, tensor: bool = True) -> npt.NDArray[np.object_ | Any]: ...

polygrid2d: Final[_FuncVal2D] = ...
polygrid3d: Final[_FuncVal3D] = ...
polyvander: Final[_FuncVander] = ...
polyvander2d: Final[_FuncVander2D] = ...
polyvander3d: Final[_FuncVander3D] = ...
polyfit: Final[_FuncFit] = ...
polycompanion: Final[_FuncCompanion] = ...
polyroots: Final[_FuncRoots] = ...

class Polynomial(ABCPolyBase[None]):
    basis_name: ClassVar[None] = None  # pyright: ignore[reportIncompatibleMethodOverride] # pyrefly: ignore[bad-override]
    domain: _Array2[np.float64 | Any] = ...  # pyright: ignore[reportIncompatibleMethodOverride]
    window: _Array2[np.float64 | Any] = ...  # pyright: ignore[reportIncompatibleMethodOverride]



"""
'Generic' Array API backend for RBF interpolation.

The general logic is this: `_rbfinterp.py` implements the user API and calls
into either `_rbfinterp_np` (the "numpy backend"), or `_rbfinterp_xp` (the
"generic backend".

The numpy backend offloads performance-critical computations to the
pythran-compiled `_rbfinterp_pythran` extension. This way, the call chain is

    _rbfinterp.py <-- _rbfinterp_np.py <-- _rbfinterp_pythran.py

The "generic" backend here is a drop-in replacement of the API of
`_rbfinterp_np.py` for use in `_rbfinterp.py` with non-numpy arrays.

The implementation closely follows `_rbfinterp_np + _rbfinterp_pythran`, with
the following differences:

  -  We used vectorized code not explicit loops in `_build_system` and
     `_build_evaluation_coefficients`; this is more torch/jax friendly;
  - RBF kernels are also "vectorized" and not scalar: they receive an
    array of norms not a single norm;
  - RBF kernels accept an extra xp= argument;

In general, we would prefer less code duplication. The main blocker ATM is
that pythran cannot compile functions with an xp= argument where xp is numpy.
"""
from numpy.linalg import LinAlgError
from ._rbfinterp_common import _monomial_powers_impl


def _monomial_powers(ndim, degree, xp):
    out = _monomial_powers_impl(ndim, degree)
    out = xp.asarray(out)
    if out.shape[0] == 0:
        out = xp.reshape(out, (0, ndim))
    return out


def _build_and_solve_system(y, d, smoothing, kernel, epsilon, powers, xp):
    """Build and solve the RBF interpolation system of equations.

    Parameters
    ----------
    y : (P, N) float ndarray
        Data point coordinates.
    d : (P, S) float ndarray
        Data values at `y`.
    smoothing : (P,) float ndarray
        Smoothing parameter for each data point.
    kernel : str
        Name of the RBF.
    epsilon : float
        Shape parameter.
    powers : (R, N) int ndarray
        The exponents for each monomial in the polynomial.

    Returns
    -------
    coeffs : (P + R, S) float ndarray
        Coefficients for each RBF and monomial.
    shift : (N,) float ndarray
        Domain shift used to create the polynomial matrix.
    scale : (N,) float ndarray
        Domain scaling used to create the polynomial matrix.

    """
    lhs, rhs, shift, scale = _build_system(
        y, d, smoothing, kernel, epsilon, powers, xp
        )
    try:
        coeffs = xp.linalg.solve(lhs, rhs)
    except Exception:
        # Best-effort attempt to emit a helpful message.
        # `_rbfinterp_np` backend gives better diagnostics; it is hard to
        # match it in a backend-agnostic way: e.g. jax emits no error at all,
        # and instead returns an array of nans for a singular `lhs`.
        msg = "Singular matrix"
        nmonos = powers.shape[0]
        if nmonos > 0:
            pmat = polynomial_matrix((y - shift)/scale, powers, xp=xp)
            rank = xp.linalg.matrix_rank(pmat)
            if rank < nmonos:
                msg = (
                    "Singular matrix. The matrix of monomials evaluated at "
                    "the data point coordinates does not have full column "
                    f"rank ({rank}/{nmonos})."
                    )
        raise LinAlgError(msg)

    return shift, scale, coeffs


def linear(r, xp):
    return -r


def thin_plate_spline(r, xp):
    # NB: changed w.r.t. pythran, vectorized
    return xp.where(r == 0, 0, r**2 * xp.log(r))


def cubic(r, xp):
    return r**3


def quintic(r, xp):
    return -r**5


def multiquadric(r, xp):
    return -xp.sqrt(r**2 + 1)


def inverse_multiquadric(r, xp):
    return 1.0 / xp.sqrt(r**2 + 1.0)


def inverse_quadratic(r, xp):
    return 1.0 / (r**2 + 1.0)


def gaussian(r, xp):
    return xp.exp(-r**2)


NAME_TO_FUNC = {
   "linear": linear,
   "thin_plate_spline": thin_plate_spline,
   "cubic": cubic,
   "quintic": quintic,
   "multiquadric": multiquadric,
   "inverse_multiquadric": inverse_multiquadric,
   "inverse_quadratic": inverse_quadratic,
   "gaussian": gaussian
   }


def kernel_matrix(x, kernel_func, xp):
    """Evaluate RBFs, with centers at `x`, at `x`."""
    return kernel_func(
        xp.linalg.vector_norm(x[None, :, :] - x[:, None, :], axis=-1), xp
    )


def polynomial_matrix(x, powers, xp):
    """Evaluate monomials, with exponents from `powers`, at `x`."""
    return xp.prod(x[:, None, :] ** powers, axis=-1)


def _build_system(y, d, smoothing, kernel, epsilon, powers, xp):
    """Build the system used to solve for the RBF interpolant coefficients.

    Parameters
    ----------
    y : (P, N) float ndarray
        Data point coordinates.
    d : (P, S) float ndarray
        Data values at `y`.
    smoothing : (P,) float ndarray
        Smoothing parameter for each data point.
    kernel : str
        Name of the RBF.
    epsilon : float
        Shape parameter.
    powers : (R, N) int ndarray
        The exponents for each monomial in the polynomial.

    Returns
    -------
    lhs : (P + R, P + R) float ndarray
        Left-hand side matrix.
    rhs : (P + R, S) float ndarray
        Right-hand side matrix.
    shift : (N,) float ndarray
        Domain shift used to create the polynomial matrix.
    scale : (N,) float ndarray
        Domain scaling used to create the polynomial matrix.

    """
    s = d.shape[1]
    r = powers.shape[0]
    kernel_func = NAME_TO_FUNC[kernel]

    # Shift and scale the polynomial domain to be between -1 and 1
    mins = xp.min(y, axis=0)
    maxs = xp.max(y, axis=0)
    shift = (maxs + mins)/2
    scale = (maxs - mins)/2
    # The scale may be zero if there is a single point or all the points have
    # the same value for some dimension. Avoid division by zero by replacing
    # zeros with ones.
    scale = xp.where(scale == 0.0, 1.0, scale)

    yeps = y*epsilon
    yhat = (y - shift)/scale

    out_kernels  = kernel_matrix(yeps, kernel_func, xp)
    out_poly = polynomial_matrix(yhat, powers, xp)

    lhs = xp.concat(
        [
         xp.concat((out_kernels, out_poly), axis=1),
         xp.concat((out_poly.T, xp.zeros((r, r))), axis=1)
        ]
    , axis=0) + xp.diag(xp.concat([smoothing, xp.zeros(r)]))

    rhs = xp.concat([d, xp.zeros((r, s))], axis=0)

    return lhs, rhs, shift, scale


def _build_evaluation_coefficients(
    x, y, kernel, epsilon, powers, shift, scale, xp
):
    """Construct the coefficients needed to evaluate
    the RBF.

    Parameters
    ----------
    x : (Q, N) float ndarray
        Evaluation point coordinates.
    y : (P, N) float ndarray
        Data point coordinates.
    kernel : str
        Name of the RBF.
    epsilon : float
        Shape parameter.
    powers : (R, N) int ndarray
        The exponents for each monomial in the polynomial.
    shift : (N,) float ndarray
        Shifts the polynomial domain for numerical stability.
    scale : (N,) float ndarray
        Scales the polynomial domain for numerical stability.

    Returns
    -------
    (Q, P + R) float ndarray

    """
    kernel_func = NAME_TO_FUNC[kernel]

    yeps = y*epsilon
    xeps = x*epsilon
    xhat = (x - shift)/scale

    # NB: changed w.r.t. pythran
    vec = xp.concat(
        [
            kernel_func(
                xp.linalg.vector_norm(
                    xeps[:, None, :] - yeps[None, :, :], axis=-1
                ), xp
            ),
            xp.prod(xhat[:, None, :] ** powers, axis=-1)
        ], axis=-1
    )

    return vec


def compute_interpolation(x, y, kernel, epsilon, powers, shift, scale, coeffs, xp):
    vec = _build_evaluation_coefficients(
        x, y, kernel, epsilon, powers, shift, scale, xp
    )
    return vec @ coeffs



# This file is not meant for public use and will be removed in SciPy v2.0.0.
# Use the `scipy.fftpack` namespace for importing the functions
# included below.

from scipy._lib.deprecation import _sub_module_deprecation

__all__ = [  # noqa: F822
    'diff',
    'tilbert', 'itilbert', 'hilbert', 'ihilbert',
    'cs_diff', 'cc_diff', 'sc_diff', 'ss_diff',
    'shift', 'convolve'
]


def __dir__():
    return __all__


def __getattr__(name):
    return _sub_module_deprecation(sub_package="fftpack", module="pseudo_diffs",
                                   private_modules=["_pseudo_diffs"], all=__all__,
                                   attribute=name)
